#define __STDC_LIMIT_MACROS
#define __STDC_FORMAT_MACROS

extern "C" {
  #include <caml/mlvalues.h>
  #include <caml/alloc.h>
  #include <caml/memory.h>
  #include <caml/custom.h>
}

//#include "minisat2/Solver.h"
#include <minisat/core/Solver.h>

//
using namespace Minisat;

extern "C" {
  CAMLprim value mkLit_stub(value var, value sign);
  CAMLprim value var_stub(value lit);
  CAMLprim value sign_stub(value lit);
  CAMLprim value neg_lit_stub(value lit);

  CAMLprim value new_solver_stub(value v);
  CAMLprim value newVar_stub(value solver);
  CAMLprim value addClause_stub(value solver, value lits);
  CAMLprim value solve_stub(value solver, value assumps);
  CAMLprim value modelValue_stub(value solver, value lit);
  CAMLprim value get_conflict_stub(value solver);

  void finalize_solver(value solver);

  char identifier[] = "minisat_stub";

  struct custom_operations solver_ops =
    { identifier, &finalize_solver, custom_compare_default,
      custom_hash_default, custom_serialize_default, custom_deserialize_default };
}

CAMLprim value mkLit_stub(value var, value sign) {
  Var v = Int_val(var);
  bool s = Bool_val(sign);
  //int l = toInt(Lit(v, s));
  int l = toInt(mkLit(v, s));
  return Val_int(l);
}

CAMLprim value var_stub(value lit) {
  Lit l = toLit(Int_val(lit));
  Var v = var(l);
  return Val_int(v);
}

CAMLprim value sign_stub(value lit) {
  Lit l = toLit(Int_val(lit));
  bool s = sign(l);
  return Val_bool(s);
}

CAMLprim value neg_lit_stub(value lit) {
  Lit l = toLit(Int_val(lit));
  int nl = toInt(~l);
  return Val_int(nl);
}

CAMLprim value new_solver_stub(value v) {
  CAMLparam1(v);
  CAMLlocal1(result);
  result = caml_alloc_custom(&solver_ops, sizeof(Solver *), 0, 1);
  Solver *solv = new Solver();
  Solver **temp = (Solver **)Data_custom_val(result);
  *temp = solv;
  CAMLreturn (result);
}

CAMLprim value newVar_stub(value solver) {
  Solver *solv = *((Solver **)Data_custom_val(solver));
  Var var = solv->newVar();
  return Val_int(var);
}

CAMLprim value addClause_stub(value solver, value lits) {
  Solver *solv = *((Solver **)Data_custom_val(solver));
  vec<Lit> ps;
  value block = lits;
  while (Is_block(block)) {
    Lit lit = toLit(Int_val(Field(block, 0)));
    ps.push(lit);
    block = Field(block, 1);
  }
  bool okay = solv->addClause(ps);
  return Val_bool(okay);
}

CAMLprim value solve_stub(value solver, value assumps) {
  Solver *solv = *((Solver **)Data_custom_val(solver));
  vec<Lit> ass;
  value block = assumps;
  while (Is_block(block)) {
    Lit lit = toLit(Int_val(Field(block, 0)));
    ass.push(lit);
    block = Field(block, 1);
  }
  bool isSat = solv->solve(ass);
  return Val_bool(isSat);
}

CAMLprim value modelValue_stub(value solver, value lit) {
  Solver *solv = *((Solver **)Data_custom_val(solver));
  Lit l = toLit(Int_val(lit));
  lbool status = solv->modelValue(l);
  return Val_int(toInt(status));
}

CAMLprim value get_conflict_stub(value solver) {
  CAMLparam1(solver);
  CAMLlocal2(res, r);
  Solver *solv = *((Solver **)Data_custom_val(solver));
  res = Val_int(0);
  for (int i = 0; i < solv->conflict.size(); ++i) {
    int lit = toInt(solv->conflict[i]);
    r = caml_alloc_small(2, 0);
    Field(r, 0) = Val_int(lit);
    Field(r, 1) = res;
    res = r;
  }
  CAMLreturn(res);
}

void finalize_solver(value solver) {
  Solver *solv = *((Solver **)Data_custom_val(solver));
  delete solv;
  return;
}
