extern "C" {
  #include <stdio.h>
  #include <readline/readline.h>
  #include <readline/history.h>
  #include <caml/mlvalues.h>
  #include <caml/alloc.h>
  #include <caml/memory.h>
  #include <caml/custom.h>
  #include <caml/callback.h>
  #include <curses.h>
}


#include <string>

extern "C" {

value set_readline_name(value name);
value readline_stub(value unit);
value readline_prompt_stub(value prompt);
value add_history_stub(value str);
value connect_to_curses(value unit);
value readline_use_ocaml_callback(value unit);
value line_buffer_stub(value unit);
value cursor_position_stub(value unit);

}

static std::string g_readline_name = "";

value set_readline_name(value name)
{
  CAMLparam1 (name);
  g_readline_name = String_val(name);
  rl_readline_name = g_readline_name.c_str();
  CAMLreturn (Val_unit);
}

value add_history_stub(value str) {
  CAMLparam1 (str);
  add_history(String_val(str));
  CAMLreturn (Val_unit);
}

// for returning an option see
// http://www.linux-nantes.org/~fmonnier/ocaml/ocaml-wrapping-c.html#ref_option

#define Val_none Val_int(0)

static value Val_some( value v )
{
    CAMLparam1( v );
    CAMLlocal1( some );
    some = caml_alloc(1, 0);
    Store_field( some, 0, v );
    CAMLreturn( some );
}

value readline_stub(value unit) {
  CAMLparam1 (unit);
  CAMLlocal1( line_data );
  char* line = readline("");
  if (line == NULL) {
    CAMLreturn( Val_none );
  } else {
    line_data = caml_alloc_string(strlen(line));
    strcpy( String_val(line_data), line);
    free(line);
    CAMLreturn( Val_some( line_data ) );
  }
}

value readline_prompt_stub(value prompt) {
  CAMLparam1 (prompt);
  CAMLlocal1( line_data );
  char* line = readline(String_val(prompt));
  if (line == NULL) {
    CAMLreturn( Val_none );
  } else {
    line_data = caml_alloc_string(strlen(line));
    strcpy( String_val(line_data), line);
    free(line);
    CAMLreturn( Val_some( line_data ) );
  }
}

// http://stackoverflow.com/questions/691652/using-gnu-readline-how-can-i-add-ncurses-in-the-same-program

int getch_stdin(FILE* f) {
    if (f == stdin) return getch();
    else return rl_getc(f);
}

value connect_to_curses(value unit) {
    CAMLparam1 (unit);
    // connect readline to ncurses according to:
    rl_getc_function = getch_stdin;
    CAMLreturn (Val_unit);
}

    // TODO rebind: rl_completion_display_matches_hook
    // TODO use: rl_line_buffer
void call_ocaml_redraw() {
    static value * closure_f = NULL;
    if (closure_f == NULL) {
        /* First time around, look up by name */
        closure_f = caml_named_value("readline_redraw_callback");
    }
    caml_callback(*closure_f, Val_unit);
}

value readline_use_ocaml_callback(value unit) {
    CAMLparam1 (unit);
    rl_redisplay_function = call_ocaml_redraw;
    CAMLreturn (Val_unit);
}

value line_buffer_stub(value unit) {
    CAMLparam1 (unit);
    CAMLlocal1( line_data );
    line_data = caml_alloc_string(strlen(rl_line_buffer)+1);
    strcpy( String_val(line_data), rl_line_buffer);
    CAMLreturn( line_data );
}

value cursor_position_stub(value unit) {
    CAMLparam1 (unit);
    CAMLreturn (Val_int (rl_point));
}


